/*
      tbb.c v1.02

      This is a double buffered copy program that works like dd(1) except
      that it uses two processes to do the copy which allows overlapped
      reading and writing. If you use tbb to copy to a streaming cartridge
      tape you can improve the streaming performance of the drive.

      tbb is written as a replacement for the tcb(1) command on A/UX 
      operating systems and so is called tbb. But it should be able to
      work on any UNIX system that implements semaphores.

      Compile it with cc -O -o tbb tbb.c

      Note: tbb uses semaphores. If you abort the program with a SIGKILL
      the program will not be able to clean up the semaphores that it
      has created. You can remove the semaphores manually with ipcrm(1)
      else they will hang around until the next reboot. It is better to
      kill the program with another signal (most users have control/C
      set to do the right kill).

      Tony Cooper
      sramtrc@albert.dsir.govt.nz

*/

#include <stdio.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/wait.h>
#include <sys/ipc.h>
#include <sys/sem.h>

/* default buffer sizes */
#define CBS	131072
#define IBS	65536
#define OBS	65536
#define PBS	512

/* number of semaphores */
#define NSEM    6

#define min(a,b) ((a)<(b) ?(a) :(b))

#define FALSE	0
#define TRUE	1

#define STDIN	0
#define STDOUT	1

int	semid[NSEM];
int	semval[NSEM] = {0,1,0,1,0,0};
int	myreadsemid, yourreadsemid, mywritesemid, yourwritesemid,
	stopreadsemid, stopwritesemid;
struct sembuf  sop;

static int
    ifd, ofd,           /* input/output file descriptors */
    ibs, obs,           /* input/output block sizes */
    bs,			/* set ibs and obs to bs */
    pbs,		/* pad block size */
    cbs,                /* internal buffer size */
    eof = FALSE,	/* end of input? */
    pid = 1;		/* set to 1 in case we abort before fork */

static char *buffer;	/* buffer pointer */
static char *buffptr;	/* points to where we are in the buffer */
static char *buffend;	/* points to end of the buffer */

extern void perror();
extern char *malloc();

void usage()
{
    fprintf(stderr,"Usage tbb [option=value]... where option is one of bs, ibs, obs, cbs, pbs, if, of\n");
    exit(1);
}

long getnum(s)
char *s;
{
    register long n;
    long strtol();
    char *p;

    n = strtol(s,&p,10);
    switch(*p) {
        case 'k':
        case 'K': n *= 1024;
                  p++;
                  break;
        case 'b':
        case 'B': n *= 512;
                  p++;
                  break;

        case 'w':
        case 'W': n *= 2;
                  p++;
                  break;

        case '\0': break;

        default: usage();
    }
    if (*p != '\0')
        usage();
    if (n < 1) {
        fprintf(stderr,"Invalid value %s\n",s);
        exit(1);
    }
    return n;
}

void parse(argc,argv)
int argc;
char *argv[];
{
    register char *s;
    register int i;

    for (i = 1; i < argc; i++) {
        s = argv[i];
        if (strncmp(s,"if=",3) == 0) {
            s += 3;
            if ((ifd = open(s, O_RDONLY),0) < 0) {
                (void)perror(s);
		exit(1);
	    }
        } else if (strncmp(s,"of=",3) == 0) {
            s += 3;
            if ((ofd = open(s, O_WRONLY | O_CREAT, 0666)) < 0) {
                (void)perror(s);
		exit(1);
	    }
        } else if (strncmp(s,"bs=",3) == 0) {
            bs = getnum(s+3);
        } else if (strncmp(s,"ibs=",4) == 0) {
            ibs = getnum(s+4);
        } else if (strncmp(s,"obs=",4) == 0) {
            obs = getnum(s+4);
        } else if (strncmp(s,"cbs=",4) == 0) {
            cbs = getnum(s+4);
        } else if (strncmp(s,"pbs=",4) == 0) {
            pbs = getnum(s+4);
        } else {
            usage();
        }
    }
    if (bs != 0) {
        ibs = bs;
        obs = bs;
    }
    if (pbs > obs) {
        pbs = obs;
    }
    if (cbs % obs) {
        fprintf(stderr,"cbs=%d is not a multiple of obs=%d\n",cbs,obs);
        exit(1);
    }
    if (cbs % ibs) {
        fprintf(stderr,"cbs=%d is not a multiple of ibs=%d\n",cbs,ibs);
        exit(1);
    }
}

void cleanup(status)
int status;
{
    /*
     * If parent, wait for the child to finish then remove the semaphores
     */
    if (pid) {
	register int i;
	union wait statloc;

	if (waitpid(pid, &statloc, 0) == -1) {
	    fprintf(stderr,"PID %d ", pid);
	    (void)perror("tbb: wait");
	} else {
	    if (status == 0) {
		if (WIFEXITED(statloc))		/* child exited */
		    status = statloc.w_retcode;
		else
		    status = 1;			/* child received signal */
	    }
	}

	for (i = 0; i < NSEM; i++) {
	    semctl(semid[i], 0, IPC_RMID, 0);
	}
    }
    exit(status);
}

void abort()
{
    /* 
     * Stop all reading and writing, wakeup the other process, then cleanup
     */
    (void)semctl(stopreadsemid, 0, SETVAL, 1);
    (void)semctl(stopwritesemid, 0, SETVAL, 1);
    sop.sem_op = 1;
    (void)semop(mywritesemid, &sop, 1);
    (void)semop(myreadsemid, &sop, 1);
    cleanup(1);
}

int stopread()
{
    register int rv;

    if ((rv = semctl(stopreadsemid, 0, GETVAL, 0)) == -1) {
	fprintf(stderr,"PID %d, semid %d, ", pid, stopreadsemid);
	(void)perror("tbb: stop read");
	abort();
    }
    return rv;
}

int stopwrite()
{
    register int rv;

    if ((rv = semctl(stopwritesemid, 0, GETVAL, 0)) == -1) {
	fprintf(stderr,"PID %d, semid %d, ", pid, stopwritesemid);
	(void)perror("tbb: stop write");
	abort();
    }
    return rv;
}

void before_read()
{
    sop.sem_op = -1;
    if (semop(yourreadsemid, &sop, 1) == -1) {
	fprintf(stderr,"PID %d, semid %d, ", pid, yourreadsemid);
	(void)perror("tbb: read semop -1");
	abort();
    }
    if (stopread()) cleanup(0);
}

void after_read()
{
    sop.sem_op = 1;
    if (semop(myreadsemid, &sop, 1) == -1) {
	fprintf(stderr,"PID %d, semid %d, ", pid, myreadsemid);
	(void)perror("tbb: read semop 1");
	abort();
    }
}

void before_write()
{
    sop.sem_op = -1;
    if (semop(yourwritesemid, &sop, 1) == -1) {
	fprintf(stderr,"PID %d, semid %d, ", pid, yourwritesemid);
	(void)perror("tbb: write semop -1");
	abort();
    }
    if (stopwrite()) cleanup(0);
}

void after_write()
{
    sop.sem_op = 1;
    if (semop(mywritesemid, &sop, 1) == -1) {
	fprintf(stderr,"PID %d, semid %d, ", pid, mywritesemid);
	(void)perror("tbb: write semop 1");
	abort();
    }
}

main(argc, argv)
int argc;
char *argv[];
{
    register int i;

    ifd = STDIN;
    ofd = STDOUT;
    cbs = CBS;
    obs = OBS;
    ibs = IBS;
    pbs = PBS;
    
    parse(argc,argv);

    for (i = 0; i < NSEM; i++) {
	if ((semid[i] = semget(IPC_PRIVATE, 1, IPC_CREAT | 0600)) == -1) {
	    (void)perror("tbb: semget");
	    abort();
	}
    }

    for (i = 0; i < NSEM; i++) {
	if (semctl(semid[i], 0, SETVAL, semval[i]) == -1) {
	    (void)perror("tbb: setval");
	    abort();
	}
    }

    sop.sem_num = 0;
    sop.sem_flg = 0;

    set42sig();
    signal(SIGINT, abort);
    signal(SIGHUP, abort);
    signal(SIGQUIT, abort);
    signal(SIGILL, abort);
    signal(SIGTRAP, abort);
    signal(SIGIOT, abort);
    signal(SIGFPE, abort);
    signal(SIGBUS, abort);
    signal(SIGSEGV, abort);
    signal(SIGSYS, abort);
    signal(SIGPIPE, abort);
    signal(SIGALRM, SIG_IGN);
    signal(SIGTERM, abort);
    signal(SIGUSR1, SIG_IGN);
    signal(SIGUSR2, SIG_IGN);
    signal(SIGPWR, abort);
    signal(SIGTTIN, abort);
    signal(SIGTTOU, abort);
    signal(SIGXCPU, abort);
    signal(SIGXFSZ, abort);
    signal(SIGVTALRM, abort);
    signal(SIGPROF, SIG_IGN);
    signal(SIGWINCH, SIG_IGN);
    signal(SIGURG, SIG_IGN);
    signal(SIGIO, SIG_IGN);

    if ((pid = fork()) == -1) {
	(void)perror("tbb: fork");
	abort();
    }

    if (pid == 0) {			/* child */
	myreadsemid =	 semid[0];
	yourreadsemid =	 semid[1];
	mywritesemid =	 semid[2];
	yourwritesemid = semid[3];
    } else {				/* parent */
	myreadsemid =	 semid[1];
	yourreadsemid =	 semid[0];
	mywritesemid =	 semid[3];
	yourwritesemid = semid[2];
    }
    stopreadsemid = semid[4];
    stopwritesemid = semid[5];

    if ((buffer = malloc((unsigned) cbs)) == NULL) {
	fprintf(stderr,"PID %d ", pid);
	(void)perror("tbb: buffer");
	abort();
    }
    buffend = buffer + cbs;

    while (!eof) {
	before_read();
	readbuffer();
	after_read();

	before_write();
	writebuffer();
	after_write();
    }
    cleanup(0);
}

readbuffer()
{
    register int bytes_wanted, bytes_actual;

    buffptr = buffer;
    while ((buffptr < buffend) && !eof) {
	bytes_wanted = min(ibs, buffend - buffptr);

	bytes_actual = read(ifd, buffptr, (unsigned)bytes_wanted);

	if (bytes_actual == 0) {       /* end of file */
	    (void)semctl(stopreadsemid, 0, SETVAL, 1);
	    close(ifd);
	    eof = TRUE;
	} else if (bytes_actual < 0) {
	    fprintf(stderr,"PID %d ", pid);
	    (void) perror("tbb: read error");
	    abort();
	}
	buffptr += bytes_actual;
    }
}

writebuffer()
{
    register int bytes_wanted, bytes_actual;
    register char *writeptr;
    register int j,k;

    writeptr = buffer;
    while (writeptr < buffptr) {
	bytes_wanted = min(obs, buffptr - writeptr);

	if ((bytes_wanted < obs) &&		/* if partial block and */
	(bytes_wanted % pbs != 0)) {		/* not multiple of pbs */
	    k = pbs - bytes_wanted % pbs;
	    for (j = 0; j < k; j++)
		*buffptr++ = '\0';		/* pad the output block */
	    bytes_wanted += k;
	}

	bytes_actual = write(ofd, writeptr, (unsigned)bytes_wanted);

	if (bytes_actual < bytes_wanted) {
	    (void)semctl(stopwritesemid, 0, SETVAL, 1);
	    close(ofd);
	    fprintf(stderr,"PID %d ", pid);
	    (void) perror("tbb: write error");
	    abort();
	}
	writeptr += bytes_actual;
    }
}
